"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Option_1 = require("fp-ts/lib/Option");
var Reader_1 = require("fp-ts/lib/Reader");
var Array_1 = require("fp-ts/lib/Array");
var ts = require("typescript");
var M = require("./model");
var monocle_ts_1 = require("monocle-ts");
exports.defaultOptions = {
    tagName: 'type',
    foldName: 'fold',
    matcheeName: 'fa',
    handlersStyle: { type: 'positional' }
};
var getLens = monocle_ts_1.Lens.fromProp();
exports.lenses = {
    tagName: getLens('tagName'),
    foldName: getLens('foldName'),
    matcheeName: getLens('matcheeName'),
    handlersStyle: getLens('handlersStyle')
};
var getMemberName = function (m, position) {
    return m.name.getOrElseL(function () { return "value" + position; });
};
var getDomainName = function (type) {
    switch (type.kind) {
        case 'TypeReference':
            return lowerCase(type.name);
        case 'TupleType':
            return 'tuple';
        case 'FunctionType':
            return 'f';
    }
};
var getType = function (type) {
    switch (type.kind) {
        case 'TypeReference':
            return ts.createTypeReferenceNode(type.name, type.parameters.map(function (p) { return getType(p); }));
        case 'TupleType':
            return ts.createTupleTypeNode([getType(type.fst), getType(type.snd)].concat(type.other.map(getType)));
        case 'FunctionType':
            return ts.createFunctionTypeNode(undefined, [getParameterDeclaration(getDomainName(type.domain), getType(type.domain))], getType(type.codomain));
    }
};
exports.data = function (d) {
    return new Reader_1.Reader(function (e) {
        var unionType = ts.createUnionTypeNode(d.constructors.toArray().map(function (c) {
            var tag = ts.createPropertySignature([ts.createModifier(ts.SyntaxKind.ReadonlyKeyword)], e.tagName, undefined, ts.createLiteralTypeNode(ts.createLiteral(c.name)), undefined);
            var members = c.members.map(function (m, position) {
                return ts.createPropertySignature([ts.createModifier(ts.SyntaxKind.ReadonlyKeyword)], getMemberName(m, position), undefined, getType(m.type), undefined);
            });
            return ts.createTypeLiteralNode([tag].concat(members));
        }));
        return ts.createTypeAliasDeclaration(undefined, [ts.createModifier(ts.SyntaxKind.ExportKeyword)], d.introduction.name, d.introduction.parameters.map(function (p) {
            return ts.createTypeParameterDeclaration(p.name, p.constraint.map(getType).toUndefined());
        }), unionType);
    });
};
var lowerCase = function (name) {
    return name.substring(0, 1).toLocaleLowerCase() + name.substring(1);
};
var getNullaryConstructorVariableStatement = function (tagName, c, introduction) {
    var name = lowerCase(c.name);
    return ts.createVariableStatement([ts.createModifier(ts.SyntaxKind.ExportKeyword)], ts.createVariableDeclarationList([
        ts.createVariableDeclaration(name, ts.createTypeReferenceNode(introduction.name, introduction.parameters.map(function (p) {
            return p.constraint
                .map(function (c) { return getType(c); })
                .getOrElse(ts.createKeywordTypeNode(ts.SyntaxKind.NeverKeyword));
        })), ts.createObjectLiteral([ts.createPropertyAssignment(tagName, ts.createStringLiteral(c.name))]))
    ], ts.NodeFlags.Const));
};
var getFunctionDeclaration = function (name, typeParameters, parameters, type, body) {
    return ts.createFunctionDeclaration(undefined, [ts.createModifier(ts.SyntaxKind.ExportKeyword)], undefined, name, typeParameters, parameters, type, body);
};
var getParameterDeclaration = function (name, type) {
    return ts.createParameter(undefined, undefined, undefined, name, undefined, type, undefined);
};
var getConstructorFunctionDeclaration = function (tagName, c, introduction) {
    var name = lowerCase(c.name);
    var typeParameters = introduction.parameters.map(function (p) {
        return ts.createTypeParameterDeclaration(p.name, p.constraint.map(getType).toUndefined());
    });
    var parameters = c.members.map(function (m, position) {
        var name = getMemberName(m, position);
        var type = getType(m.type);
        return getParameterDeclaration(name, type);
    });
    var type = ts.createTypeReferenceNode(introduction.name, introduction.parameters.map(function (p) { return ts.createTypeReferenceNode(p.name, []); }));
    var body = ts.createBlock([
        ts.createReturn(ts.createObjectLiteral([
            ts.createPropertyAssignment(tagName, ts.createStringLiteral(c.name))
        ].concat(c.members.map(function (m, position) {
            var name = getMemberName(m, position);
            return ts.createShorthandPropertyAssignment(name);
        }))))
    ]);
    return getFunctionDeclaration(name, typeParameters, parameters, type, body);
};
exports.constructors = function (d) {
    return new Reader_1.Reader(function (e) {
        return d.constructors.toArray().map(function (c) {
            if (c.members.length === 0) {
                return getNullaryConstructorVariableStatement(e.tagName, c, d.introduction);
            }
            else {
                return getConstructorFunctionDeclaration(e.tagName, c, d.introduction);
            }
        });
    });
};
var isNullaryConstructor = function (c) {
    return c.members.length === 0;
};
var admitsEagerFold = function (d) {
    return d.constructors.toArray().some(isNullaryConstructor);
};
var isSumType = function (d) {
    return d.constructors.length() > 1;
};
var getFoldReturnTypeParameterName = function (i) {
    var base = 'R';
    var candidate = base;
    var counter = 0;
    while (i.parameters.findIndex(function (_a) {
        var name = _a.name;
        return candidate === name;
    }) !== -1) {
        candidate = base + ++counter;
    }
    return candidate;
};
var getHandlerName = function (c) {
    return "on" + c.name;
};
var getPositionalFoldHandlers = function (d, isEager) {
    var returnTypeParameterName = getFoldReturnTypeParameterName(d.introduction);
    return d.constructors.toArray().map(function (c) {
        var type = isEager && isNullaryConstructor(c)
            ? ts.createTypeReferenceNode(returnTypeParameterName, [])
            : ts.createFunctionTypeNode(undefined, c.members.map(function (m, position) { return getParameterDeclaration(getMemberName(m, position), getType(m.type)); }), ts.createTypeReferenceNode(returnTypeParameterName, []));
        return getParameterDeclaration(getHandlerName(c), type);
    });
};
var getRecordFoldHandlers = function (d, handlersName, isEager) {
    var returnTypeParameterName = getFoldReturnTypeParameterName(d.introduction);
    var type = ts.createTypeLiteralNode(d.constructors.toArray().map(function (c) {
        var type = isEager && isNullaryConstructor(c)
            ? ts.createTypeReferenceNode(returnTypeParameterName, [])
            : ts.createFunctionTypeNode(undefined, c.members.map(function (m, position) { return getParameterDeclaration(getMemberName(m, position), getType(m.type)); }), ts.createTypeReferenceNode(returnTypeParameterName, []));
        return ts.createPropertySignature(undefined, getHandlerName(c), undefined, type, undefined);
    }));
    return [getParameterDeclaration(handlersName, type)];
};
var getPositionalFoldBody = function (d, matcheeName, tagName, isEager) {
    return ts.createBlock([
        ts.createSwitch(ts.createPropertyAccess(ts.createIdentifier(matcheeName), tagName), ts.createCaseBlock(d.constructors.toArray().map(function (c) {
            var access = ts.createIdentifier(getHandlerName(c));
            return ts.createCaseClause(ts.createStringLiteral(c.name), [
                ts.createReturn(isEager && isNullaryConstructor(c)
                    ? access
                    : ts.createCall(access, [], c.members.map(function (m, position) {
                        return ts.createPropertyAccess(ts.createIdentifier(matcheeName), getMemberName(m, position));
                    })))
            ]);
        })))
    ]);
};
var getRecordFoldBody = function (d, matcheeName, tagName, handlersName, isEager) {
    return ts.createBlock([
        ts.createSwitch(ts.createPropertyAccess(ts.createIdentifier(matcheeName), tagName), ts.createCaseBlock(d.constructors.toArray().map(function (c) {
            var access = ts.createPropertyAccess(ts.createIdentifier(handlersName), getHandlerName(c));
            return ts.createCaseClause(ts.createStringLiteral(c.name), [
                ts.createReturn(isEager && isNullaryConstructor(c)
                    ? access
                    : ts.createCall(access, [], c.members.map(function (m, position) {
                        return ts.createPropertyAccess(ts.createIdentifier(matcheeName), getMemberName(m, position));
                    })))
            ]);
        })))
    ]);
};
var getFold = function (d, name, isEager) {
    return new Reader_1.Reader(function (e) {
        var matcheeName = e.matcheeName;
        var tagName = e.tagName;
        var handlersStyle = e.handlersStyle;
        var returnTypeParameterName = getFoldReturnTypeParameterName(d.introduction);
        var typeParameters = d.introduction.parameters
            .concat(M.parameter(returnTypeParameterName, Option_1.none))
            .map(function (p) { return ts.createTypeParameterDeclaration(p.name, p.constraint.map(getType).toUndefined()); });
        var matchee = getParameterDeclaration(matcheeName, ts.createTypeReferenceNode(d.introduction.name, d.introduction.parameters.map(function (p) { return ts.createTypeReferenceNode(p.name, []); })));
        var handlers = handlersStyle.type === 'positional'
            ? getPositionalFoldHandlers(d, isEager)
            : getRecordFoldHandlers(d, handlersStyle.handlersName, isEager);
        var parameters = [matchee].concat(handlers);
        var type = ts.createTypeReferenceNode(returnTypeParameterName, []);
        var body = handlersStyle.type === 'positional'
            ? getPositionalFoldBody(d, matcheeName, tagName, isEager)
            : getRecordFoldBody(d, matcheeName, tagName, handlersStyle.handlersName, isEager);
        return getFunctionDeclaration(name, typeParameters, parameters, type, body);
    });
};
exports.fold = function (d) {
    return Reader_1.ask().chain(function (e) {
        var folds = Array_1.empty;
        if (isSumType(d)) {
            if (admitsEagerFold(d)) {
                folds = [getFold(d, e.foldName, true), getFold(d, e.foldName + "L", false)];
            }
            else {
                folds = [getFold(d, e.foldName, false)];
            }
        }
        return Array_1.array.sequence(Reader_1.reader)(folds);
    });
};
