import { NonEmptyArray } from 'fp-ts/lib/NonEmptyArray';
import { Option } from 'fp-ts/lib/Option';
export declare type Identifier = string;
interface TypeReference {
    readonly kind: 'TypeReference';
    readonly name: Identifier;
    readonly parameters: Array<Type>;
}
interface TupleType {
    readonly kind: 'TupleType';
    readonly fst: Type;
    readonly snd: Type;
    readonly other: Array<Type>;
}
interface FunctionType {
    readonly kind: 'FunctionType';
    readonly domain: Type;
    readonly codomain: Type;
}
export declare type Type = TypeReference | TupleType | FunctionType;
export interface Member {
    readonly type: Type;
    readonly name: Option<Identifier>;
}
export interface Constructor {
    readonly name: Identifier;
    readonly members: Array<Member>;
}
export interface Parameter {
    readonly name: Identifier;
    readonly constraint: Option<Type>;
}
export interface Introduction {
    readonly name: Identifier;
    readonly parameters: Array<Parameter>;
}
export interface Data {
    readonly introduction: Introduction;
    readonly constructors: NonEmptyArray<Constructor>;
}
export declare const typeReference: (name: string, parameters?: Type[]) => Type;
export declare const tupleType: (fst: Type, snd: Type, other?: Type[]) => Type;
export declare const functionType: (domain: Type, codomain: Type) => Type;
export declare const member: (type: Type, name?: Option<string>) => Member;
export declare const constructor: (name: string, members?: Member[]) => Constructor;
export declare const parameter: (name: string, constraint?: Option<Type>) => Parameter;
export declare const introduction: (name: string, parameters?: Parameter[]) => Introduction;
export declare const data: (introduction: Introduction, head: Constructor, tail?: Constructor[]) => Data;
export {};
