"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var P = require("parser-ts");
var S = require("parser-ts/lib/string");
var C = require("parser-ts/lib/char");
var M = require("./model");
var Option_1 = require("fp-ts/lib/Option");
var function_1 = require("fp-ts/lib/function");
var isDigit = function (c) { return '0123456789'.indexOf(c) !== -1; };
var isPunctuation = function (c) { return '| =\n():,{};[]->'.indexOf(c) !== -1; };
var identifierFirstLetter = P.sat(function (c) { return !isDigit(c) && !isPunctuation(c); });
var identifierBody = P.sat(function (c) { return !isPunctuation(c); });
var expected = function (message, parser) {
    return P.expectedL(parser, function (remaining) { return "Expected " + message + ", cannot parse " + JSON.stringify(remaining); });
};
exports.identifier = expected('an identifier', P.fold([identifierFirstLetter, C.many(identifierBody)]));
var leftParens = P.fold([C.char('('), S.spaces]);
var rightParens = P.fold([S.spaces, C.char(')')]);
var withParens = function (parser) {
    return leftParens.applySecond(parser).applyFirst(rightParens);
};
var typeReferenceLeaf = exports.identifier.map(function (name) { return M.typeReference(name); });
exports.typeReference = exports.identifier.chain(function (name) {
    return S.spaces.applySecond(exports.types.map(function (parameters) { return M.typeReference(name, parameters); }));
});
var comma = P.fold([S.spaces, C.char(','), S.spaces]);
var other = comma.chain(function () { return P.sepBy(comma, exports.type); }).alt(P.parser.of([]));
exports.tupleType = expected('a tuple', leftParens
    .chain(function () {
    return exports.type.applyFirst(comma).chain(function (fst) { return exports.type.chain(function (snd) { return other.map(function (other) { return M.tupleType(fst, snd, other); }); }); });
})
    .applyFirst(rightParens));
var arrow = P.fold([S.spaces, S.string('->'), S.spaces]);
exports.functionType = expected('a function type', S.spaces.chain(function () {
    return withParens(exports.functionType)
        .alt(exports.typeReference)
        .alt(exports.tupleType)
        .chain(function (domain) { return arrow.applySecond(exports.type).map(function (codomain) { return M.functionType(domain, codomain); }); });
}));
exports.type = exports.functionType.alt(exports.typeReference).alt(exports.tupleType);
exports.types = P.sepBy(S.spaces, exports.functionType
    .alt(typeReferenceLeaf)
    .alt(withParens(exports.typeReference))
    .alt(exports.tupleType));
var pair = exports.identifier.chain(function (name) {
    return P.fold([S.spaces, S.string('::'), S.spaces])
        .applySecond(exports.type)
        .map(function (type) { return function_1.tuple(name, type); });
});
var objectType = P.fold([C.char('{'), S.spaces])
    .applySecond(P.sepBy(comma, pair))
    .applyFirst(P.fold([S.spaces, C.char('}')]));
var recordConstructor = exports.identifier.chain(function (name) {
    return S.spaces.applySecond(objectType.map(function (pairs) { return M.constructor(name, pairs.map(function (_a) {
        var name = _a[0], type = _a[1];
        return M.member(type, Option_1.some(name));
    })); }));
});
var positionalConstructor = exports.identifier.chain(function (name) {
    return S.spaces.applySecond(exports.types.map(function (types) { return M.constructor(name, types.map(function (type) { return M.member(type); })); }));
});
exports.constructor = recordConstructor.alt(positionalConstructor);
var equal = P.fold([S.spaces, C.char('='), S.spaces]);
var parameterWithoutConstraint = exports.identifier.map(function (name) { return M.parameter(name); });
var parameterWithConstraint = P.fold([C.char('('), S.spaces]).applySecond(pair.map(function (_a) {
    var name = _a[0], type = _a[1];
    return M.parameter(name, Option_1.some(type));
}).applyFirst(P.fold([S.spaces, C.char(')')])));
exports.parameter = expected('a parameter', parameterWithoutConstraint.alt(parameterWithConstraint));
exports.introduction = expected('a data type declaration', S.string('data').chain(function () {
    return S.spaces.applySecond(exports.identifier.chain(function (name) {
        return S.spaces
            .applySecond(P.sepBy(S.spaces, exports.parameter))
            .map(function (parameters) { return M.introduction(name, parameters); })
            .applyFirst(equal);
    }));
}));
var pipe = P.fold([S.spaces, C.char('|'), S.spaces]);
exports.data = expected('a data declaration', exports.introduction
    .chain(function (definition) {
    return P.sepBy1(pipe, exports.constructor).map(function (constructors) { return M.data(definition, constructors.head, constructors.tail); });
})
    .applyFirst(S.spaces)
    .applyFirst(P.eof));
exports.parse = function (s) {
    return exports.data.run(s).bimap(function (e) { return e.message; }, function (_a) {
        var data = _a[0];
        return data;
    });
};
